% DSA-BD: main SDP solver simple.
%%
%% Explained in detail in <a title="Software_files/DSA_BD_Quick_Guide.pdf" href="http://www2.isye.gatech.edu/~cod3/CamiloOrtiz/Software_files/DSA_BD_Quick_Guide.pdf" class="style_1">Quick user guide</a>
%%
%% Solves problems in the standard primal and dual SDP of the form:
%%
%% (P) min{ <C,X> : A(X) = b, X positive semidefinite}
%% (D) min{ <b,y> : At(y) + Z = C, Z positive semidefinite}
%%
%%*************************************************************************
%% C.Ortiz
%% Last Modified : 3/31/2011
function [obj,X,y,Z,runhist] = DSA_BD(blk,At,C,b,par,X0,y0,Z0)


if isfield(par,'tol'); tol = par.tol;else tol = 1e-3; end;
sigma   = par.sigma;
sigma_2 = sigma^2;



y = y0;
if iscell(Z0); Z = Z0; else Z = {Z0}; end;
if iscell(X0); X = X0; else X = {X0}; end;



m = length(b);

fprintf('\n\n ****Begin DSA_BD algorithm. iterations = %d ****\n', par.maxit);
fprintf(DSA_BD_LegendAppend(par));
fprintf([' n = ',num2str(par.n),', m = ',num2str(par.m),'\n']);

% tstart = cputime;
tstart = clock;
%%-----------------------------------------

normb = norm(b);
normC = DSA_BD_ops(C,'norm');%norm(C)



%% normalize =  1:divide Ax=b by normb and A*y+z=c by normC
%               2:Precondition A  by L inverse where LtL = AAt
%               3-Precondition A  by SL inverse where SLtLSt = AAt
normalize = par.normalize;

%% adaptivelambda   0: no agressive choice of extragradient step size
%                   1: agressive choice of extragradient step size
adaptivelambda = par.adaptivelambda;

%% ad_iteration:    the frequency in which the adaptive agressive choice of
%                   extragradient step size
ad_iteration = par.adaptiveiteration;

%% dynamicscaling = 1: dynamic scaling of the primal inner product (dynamic change of the weight in primal and dual intermediate stepsizes)
%                   0: no dynamic scaling of the primal inner product
dynamicscaling = par.dynamicscaling;

%% dyn_scale_update_it: # of iterations to update the primal inner product
%                       scaling
dyn_scale_update_it = par.dyn_scale_updateiteration;

%% scaleratio: accepted ratio between primal and dual relative residuals
%              for of the dynamic inner product scaling
scaleratio = par.scaleratio;

%% scalecorrection: change made in the primal inner product if the primal
%                   and dual relative residuals are not within the accepted
%                   ratio for the dynamic inner product scaling
scalecorrection = par.scalecorrection;

%% scalerelnorm: 0: Checks the ratio of the primal and dual relative residuals
%                   before normalizing for the dynamic inner product scaling
%                1: Checks the ratio of the primal and dual relative residuals
%                   after normalizing for the dynamic inner product scaling
scalerelnorm = par.scalerelnorm;

%% balancedinit =   0: No balanced initialization
%                   1: Set y0 = argmin norm(Ainv(U)y-c)^2 where U=AAt.
%                   2: Choose theta such that the primal and dual relative
%                      residuals after one iterations are O(1).
%                   3: Perform 1 and 2
balancedinit = par.balancedinit;

%% Normalize the problem
L.R = sparse(m,m);
L.Rt = L.R';
L.S = sparse(m,m);
L.St = L.S';
b_orig = b;
if normalize  == 1
    b = b/(normb+1);
    C = DSA_BD_ops(C,'/',(normC+1));%C/(normC+1)
    %A = DSA_BD_ops(A, '/',(normb+1));%A /(normb+1);
    At = DSA_BD_ops(At,'/',(normb+1));%At/(normb+1);
    %     for i = 1:m
    %         At{i} = At{i}/(normb+1);
    %     end
    y=y0*(normC+1)/(normb+1);
end
if normalize  >= 2 || balancedinit > 1
    [L] = DSA_BD_cholAAt(blk,At,m);
    runhist.nnz_L_A = nnz(L.R);
    fprintf('# non-zeros of chol L_A: %d\n',runhist.nnz_L_A);
    if L.p > 0
        fprintf(' rows of A linearly dependent. p = %4.0d, m = %4.0d\n',L.p,m);
        error(' nothing done');
    end
    if normalize >= 2
        b  = L.Rt\(L.St* b);%b  = L_A\ b;
    end
end
%% *************************************************************************
% Calculate the Lipschitz Constant and lambda
% -----------------------------------------
%
Lip       = DSA_BD_computeLipschitzConstant(blk,At,par);%computeLipschitzConstant3(A,At,n,par);


lambda = sigma/Lip;



%% *************************************************************************
%  Balance the initial iterates and initial primal inner product scaling
% -----------------------------------------
%
if balancedinit == 0
    theta = 1;
else
    if balancedinit ~= 1
        theta = 1;
        if normalize < 2
            y = (L.S*(L.R\(L.Rt\(L.St*DSA_BD_AXfun(blk,At,C)))));
        elseif normalize >= 2
            y = (L.Rt\(L.St*DSA_BD_AXfun(blk,At,C)));%-(L_A\(St*AXfun3(A,C)));
        end
        X = DSA_BD_ops(X,'zeros');
        if ~iscell(X); X = {X}; end;
    end
    if balancedinit ~= 2
        theta = 1;
        AX_prev = DSA_BD_AXfun(blk,At,X,normalize,L);%AXfun3(A,X,normalize,L_A,St);
        y_prev = y;
        X_prev = X;
        lambda_x = lambda/theta;
        lambda_y = lambda*theta;
        [y,X,y_tilde,X_tilde,AX_tilde,AX_tilde_ORIG,Aty_tilde] = ...
            PerformIteration(blk,y_prev,X_prev,AX_prev,lambda_y,lambda_x,lambda,theta,sigma_2,C,b,At,L, ...
            normalize,adaptivelambda,1,ad_iteration,par);
        [rel_norm1,rel_norm2] ...
            = ComputeNormOfResuduals(blk,X_prev,X_tilde,lambda_x,C,Aty_tilde,AX_tilde,normb,normC,normalize,b,b_orig,AX_tilde_ORIG);

        balcondition = 1;
        if rel_norm1 > rel_norm2 && balancedinit == 4
            balcondition = 2;
        elseif  balancedinit ~= 4 && max(rel_norm1,1) > rel_norm2
            balcondition = 0;
        end


        while balcondition >0
            if balcondition == 1
                theta = 2*theta;
            else
                theta = theta/2;
            end
            lambda_x = lambda/theta;
            lambda_y = lambda*theta;
            [y,X,y_tilde,X_tilde,AX_tilde,AX_tilde_ORIG,Aty_tilde] = ...
                PerformIteration(blk,y_prev,X_prev,AX_prev,lambda_y,lambda_x,lambda,theta,sigma_2,C,b,At,L, ...
                normalize,adaptivelambda,1,ad_iteration,par);
            [rel_norm1,rel_norm2] ...
                = ComputeNormOfResuduals(blk,X_prev,X_tilde,lambda_x,C,Aty_tilde,AX_tilde,normb,normC,normalize,b,b_orig,AX_tilde_ORIG);
            if balcondition == 1  && max(rel_norm1,1) < rel_norm2
                balcondition = 1;
            elseif balcondition == 1 && balancedinit == 4 && rel_norm1 < rel_norm2
                balcondition = 1;
            elseif balcondition == 2 && rel_norm1 > rel_norm2
                balcondition = 2;
            else
                balcondition = 0;
            end
        end
        X = X_prev;
        y = y_prev;
    end
end
lambda_x = lambda/theta;
lambda_y = lambda*theta;

%%
if adaptivelambda == 1
    lambda_a = lambda;
end

objbest = [DSA_BD_trace(blk,C,X), b'*y,0,0];%[sum(sum(C.*X)), b'*y,0,0];

stopCritSatisfied = 0;
%%
if par.print == 1
    fprintf(' it     secs   dual          primal      lg(rrd)    lg(rrp)   sigma    theta   lg(rrd_NORM) lg(rrp_NORM)\n');
end

% runhist.preprocesstime = cputime - tstart;
runhist.preprocesstime = etime(clock,tstart);
% tstart = cputime;
tstart = clock;

%%
for iter = 1:par.maxit;
    %% Save previous variables
    y_prev = y;
    if adaptivelambda == 0 && iter > 1
        X_prev = X_tilde;
    else
        X_prev = X;
    end

    if iter > 1
        if adaptivelambda == 0
            AX_prev = AX_tilde;
        else
            AX_prev = (1-lambda_a/lambda)*AX_prev+(lambda_a/lambda)*AX_tilde;
        end
    else
        AX_prev = DSA_BD_AXfun(blk,At,X_prev,par.normalize,L);
    end
    %% Perform main iteration

    [y,X,y_tilde,X_tilde,AX_tilde,AX_tilde_ORIG,Aty_tilde,lambda_a] = ...
        PerformIteration(blk,y_prev,X_prev,AX_prev,lambda_y,lambda_x,lambda,theta,sigma_2,C,b,At,L, ...
        normalize,adaptivelambda,iter,ad_iteration,par);
    %% Obtain values of residuals


        [rel_norm1,rel_norm2,resNorm,eps,rel_norm1_ORIG,rel_norm2_ORIG,Z_tilde] ...
            = ComputeNormOfResuduals(blk,X_prev,X_tilde,lambda_x,C,Aty_tilde,AX_tilde,normb,normC,normalize,b,b_orig,AX_tilde_ORIG);



        %% Compute feasibility
            pfeas = rel_norm2;
            dfeas = rel_norm1;
            
            %% compute primal and dual objective.
            %---------------------------------------------------------------
            % Compute Obj Values

            obj = [DSA_BD_trace(blk,C,X_tilde),b'*y_tilde];%[sum(sum(C.*X_tilde)),-b'*y_tilde];

            if normalize == 1
                obj = obj*(normC+1);
            end

            secs = etime(clock,tstart);
            %% check stopping criteria
            if max(rel_norm1, rel_norm2) < tol
                fprintf( '%3.0d %8.2f %13.5e %13.5e %8.3f %8.3f  %9.6f %9.6f %8.3f %8.3f\n', iter, ...
                    secs, full(obj(2)), full(obj(1)), full(-log10(rel_norm1)), full(-log10(rel_norm2)), sigma, theta, full(-log10(rel_norm1_ORIG)), full(-log10(rel_norm2_ORIG)));
                stopCritSatisfied = 1;
            end


            %% print current progress
            if par.print == 1 &&  (iter == 1 || (mod(iter,10)==0)) && stopCritSatisfied == 0;

                fprintf( '%3.0d %8.2f %13.5e %13.5e %8.3f %8.3f  %9.6f %9.6f %8.3f %8.3f\n', iter, ...
                    secs, full(obj(2)), full(obj(1)), full(-log10(rel_norm1)), full(-log10(rel_norm2)), sigma, theta, full(-log10(rel_norm1_ORIG)), full(-log10(rel_norm2_ORIG)));

            end

            %% Calculate best so far
            if iter == 1 || bestNorm >= resNorm
                bestNorm = resNorm;
                objbest(1)=obj(1);
                objbest(2)=obj(2);
                objbest(3)=iter;
            end

            %% Dynamic scaling of the inner product
            if scalerelnorm == 1
                wrel_norm2 = rel_norm2_ORIG;
                wrel_norm1 = rel_norm1_ORIG;
            else
                wrel_norm2 = rel_norm2;
                wrel_norm1 = rel_norm1;
            end
            if mod(iter,dyn_scale_update_it)==0 && dynamicscaling == 1
                if  wrel_norm2/wrel_norm1 > scaleratio
                    theta = theta/scalecorrection;
                elseif  wrel_norm1/wrel_norm2 > scaleratio
                    theta = theta*scalecorrection;

                end
                lambda_x = lambda/theta;
                lambda_y = lambda*theta;
            end
            %% Save history
            runhist.pobj(iter)  = obj(1);
            runhist.dobj(iter)  = obj(2);
            runhist.pobjbest(iter)  = objbest(1);
            runhist.dobjbest(iter)  = objbest(2);

            runhist.resnorm(iter) = resNorm;

            runhist.epsilon(iter) = eps;
            runhist.plambda(iter)  = lambda_x;
            runhist.dlambda(iter)  = lambda_y;
            runhist.cputime(iter) = secs+runhist.preprocesstime;


            runhist.bestresnorm(iter) = bestNorm;

            if adaptivelambda == 1 && mod(iter,ad_iteration) == 0
                runhist.lambdaratio(iter) = (lambda_a)/(lambda);
            end


            runhist.pfeas(iter) = pfeas;%norm(AXfun3(A,X_tilde,par.normalize,L_A,St) - b);
            runhist.dfeas(iter) = dfeas;


            %% stop if criteria satisfied
            if stopCritSatisfied || iter>=par.maxit
                runhist. maxitreached = 0;
                if (iter>=par.maxit);
                    fprintf('max iterations reached. \n');
                    runhist. maxitreached = 1;
                end
                X = X_tilde;
                if normalize ==1
                    y = y_tilde*(normC+1)/(1+normb+1);
                elseif normalize >= 2
                    y = L.S*(L.R\y_tilde);
                else
                    y = y_tilde;
                end

                runhist.processtime = secs;
                runhist.numiterations = iter;
                runhist.eachiteration = secs/iter;
                runhist.totaltime = secs+runhist.preprocesstime;
                runhist.finalbestresnorm = bestNorm;
                runhist.bestiteration  = objbest(3);

                fprintf('preprocess time: %10.3f \n',runhist.preprocesstime);
                fprintf('# iterations: %d \n',runhist.numiterations);
                fprintf('iterations time: %10.3f \n',runhist.processtime);
                fprintf('each iteration time: %f \n',runhist.eachiteration);
                fprintf('Resulting -LOG(norm): %f \n',-log10(runhist.finalbestresnorm));
                fprintf('total time: %10.3f \n',runhist.totaltime);
                return
            end

end

fprintf('\n\n ****End DSA-BD algorithm ****\n');
%%---------------------------------------------------------------
%% end of main loop
%%---------------------------------------------------------------
%%
%% *****************************************************************************

function [y,X,y_tilde,X_tilde,AX_tilde,AX_tilde_ORIG,Aty_tilde,lambda_a] = ...
    PerformIteration(blk,y_prev,X_prev,AX_prev,lambda_y,lambda_x,lambda,theta,sigma_2,C,b,At,L, ...
    normalize,adaptivelambda,iter,ad_iteration,par)
%% \tilde y_k = y_{k-1} +  \lambda(A x_{k-1} - b)
y_tilde = y_prev-lambda_y*(AX_prev - b);

%% \tilde x_{k} = P_X(x_{k-1} - \lambda(\nabla f(x_{k-1}) - A^*\tilde y_{k}))
Aty_tilde = DSA_BD_Atyfun(blk,At,y_tilde,par.normalize,L);
ProjArg = DSA_BD_ops(X_prev,'calc3',C,lambda_x, Aty_tilde);
X_tilde = project(blk,ProjArg);



%% Get valbues of AX without normalizing if using L inverse (normaliz>=2)
if normalize < 2
    AX_tilde = DSA_BD_AXfun(blk,At,X_tilde,par.normalize,L);
else
    [AX_tilde,AX_tilde_ORIG] = DSA_BD_AXfun(blk,At,X_tilde,par.normalize,L);
end

%% Perform aggresive choice of extragradient stepsize
if adaptivelambda == 0 || mod(iter,ad_iteration) ~= 0 || iter == 1

    X = X_tilde;

    y = y_tilde - lambda_y*(AX_tilde - AX_prev);

    lambda_a = lambda;
else
    dif_A_x = (AX_tilde - AX_prev);

    v_x_norm = DSA_BD_ops(X_prev,'diffnorm',X_tilde);

    v_y = AX_tilde-b;

    gamma_k = lambda_y*((v_y)'*dif_A_x);

    eta_k = (v_x_norm/lambda)^2 + (norm(v_y)^2);

    nu_k = sigma_2*( (theta*v_x_norm)^2 + norm(y_tilde-y_prev)^2) - (lambda_y*norm(dif_A_x))^2;

    if nu_k < 0
        fprintf('discriminant = %f \n', nu_k);
        error('\|a\|^2-\sigma^2\|\tilde x_k - x_{k-1}\|^2 > 0');
    end
    discriminant = sqrt(gamma_k^2+ eta_k*nu_k);
    if gamma_k <= 0
        beta = (-gamma_k + discriminant)/eta_k;
    else
        beta = nu_k/(gamma_k+discriminant);
    end
    lambda_a_y = lambda_y + beta;
    lambda_a   = lambda_a_y /theta;

    X = DSA_BD_ops(X_prev,'calc7',lambda_a,lambda,X_tilde);%X = (1-lambda_a/lambda)X_prev+(lambda_a/lambda)X_tilde;
    y = y_prev-lambda_a_y*v_y;



end

function [rel_norm1,rel_norm2,resNorm,eps,rel_norm1_ORIG,rel_norm2_ORIG,p] ...
    = ComputeNormOfResuduals(blk,X_prev,X_tilde,lambda_x,C,Aty_tilde,AX_tilde,normb,normC,normalize,b,b_orig,AX_tilde_ORIG)
%% compute p_k^x
p = DSA_BD_ops(X_tilde,'calc1',X_prev,lambda_x, C,Aty_tilde);


%% Compute residual vector From Theorem 5.3

res1 = DSA_BD_ops(C,'calc2',Aty_tilde,p);%Aty_tilde-C+p =-C+ Aty_tilde+p %DSA_BD_ops(DSA_BD_ops(C,'+',Aty_tilde),'+',p);%(C+Aty_tilde)+p;

%res1 = C+Atyfun3(At,y_tilde,n,par.normalize,L_At,S)+p;
if normalize < 2
    res2 = b-AX_tilde;

    norm2 = norm(res2);
    norm2_ORIG = norm2;
    %res2 = b-AXfun3(A,X_tilde,par.normalize,L_A,St);
else
    res2 = b_orig - AX_tilde_ORIG;

    norm2 = norm(res2);
    res2_ORIG = b-AX_tilde;
    norm2_ORIG = norm(res2_ORIG);
    %res2 = b_orig-AXfun3(A,X_tilde);
end
%% Compute the norm of the residual
norm1 = DSA_BD_ops(res1,'norm');
%         norm2 = norm(res2);


%% complementarity in pointwise
eps = DSA_BD_trace(blk,X_tilde,p);%sum(sum(X_tilde.*p));
if normalize ~= 1
    rel_norm1 = norm1/(1+normC);
    rel_norm2 = norm2/(1+normb);
    rel_norm1_ORIG = rel_norm1;
    rel_norm2_ORIG = norm2_ORIG/(1+normb);
else
    rel_norm1 = norm1;
    rel_norm2 = norm2;
    rel_norm1_ORIG = norm1/(1+normC);
    rel_norm2_ORIG = norm2_ORIG/(1+normb);
end;
resNorm = sqrt(rel_norm1^2 + rel_norm2^2);
