% SDP_Solver: solve an semidefinite-linear program with DSA-BD method.
%%
%%  [obj,X,y,Z,runhist] = SDP_Solver(blk,At,C,b,OPTIONS,X0,y0,Z0);
%%
%%  Input: blk: a cell array describing the block diagonal structure.
%%          At: a cell array with At{p} = [svec(Ap1) ... svec(Apm)]
%%         b,C: data for the SDP instance.
%%  (X0,y0,Z0): an initial iterate
%%     OPTIONS: a structure that specifies parameters required in DSA_BD.m,
%%              (if it is not given, the default in DSA_BD_parameters.m is used).
%%
%%  Output: obj  = [<C,X> <b,y>].
%%          (X,y,Z): an approximately optimal solution or a primal or dual
%%                   infeasibility certificate.
%%          runhist: values measuring the progress of the algorithm
%%-------------------------------------------------------------------------
%%*************************************************************************
%% C. Ortiz
%% Last Modified: 3/31/2011
%% Last Modified: 11/27/2011
%%*************************************************************************

function [obj,X,y,Z,runhist] = DSA_BD_SDP_Solver(blk,At,C,b,OPTIONS,X0,y0,Z0)

%%
%%-----------------------------------------
%% get parameters from the OPTIONS structure.
%%-----------------------------------------
%%
warning off;

matlabversion = sscanf(version,'%f');

par = DSA_BD_Parameters;
if strcmp(computer,'PCWIN64') | strcmp(computer,'GLNXA64')
    par.computer = 64;
else
    par.computer = 32;
end
par.matlabversion = matlabversion(1);
par.tstart = clock;

%%
if exist('OPTIONS')
    if isfield(OPTIONS,'vers');                           par.vers                        = OPTIONS.vers; end
    if isfield(OPTIONS,'maxit');                          par.maxit                       = OPTIONS.maxit; end
    if isfield(OPTIONS,'sigma');                          par.sigma                       = OPTIONS.sigma; end
    if isfield(OPTIONS,'tol');                            par.tol                         = OPTIONS.tol; end
    if isfield(OPTIONS,'normalize');                      par.normalize                   = OPTIONS.normalize; end
    if isfield(OPTIONS,'adaptivelambda');                 par.adaptivelambda              = OPTIONS.adaptivelambda; end
    if isfield(OPTIONS,'dynamicscaling');                 par.dynamicscaling              = OPTIONS.dynamicscaling; end
    if isfield(OPTIONS,'scaleratio');                     par.scaleratio                  = OPTIONS.scaleratio; end
    if isfield(OPTIONS,'scalecorrection');                par.scalecorrection             = OPTIONS.scalecorrection; end
    if isfield(OPTIONS,'dyn_scale_updateiteration');      par.dyn_scale_updateiteration   = OPTIONS.dyn_scale_updateiteration; end
    if isfield(OPTIONS,'adaptiveiteration');              par.adaptiveiteration           = OPTIONS.adaptiveiteration; end
    if isfield(OPTIONS,'print');                          par.print                       = OPTIONS.print; end
    if isfield(OPTIONS,'n');                              par.n                           = OPTIONS.n; end
    if isfield(OPTIONS,'m');                              par.m                           = OPTIONS.m; end
    if isfield(OPTIONS,'balancedinit');                   par.balancedinit                = OPTIONS.balancedinit; end
    if isfield(OPTIONS,'file');                           par.file                        = OPTIONS.file; end
    if isfield(OPTIONS,'tstart');                         par.tstart                      = OPTIONS.tstart; end
    if isfield(OPTIONS,'scalerelnorm');                   par.scalerelnorm                = OPTIONS.scalerelnorm; end
end
if (size(blk,2) > 2); par.smallblkdim = 0; end



%%-----------------------------------------
%% convert matrices to cell arrays.
%%-----------------------------------------
%%
if ~iscell(At); At = {At}; end;
if ~iscell(C);  C = {C}; end;
if all(size(At) == [size(blk,1), length(b)]);
    convertyes = zeros(size(blk,1),1);
    for p = 1:size(blk,1)
        if strcmp(blk{p,1},'s') && all(size(At{p,1}) == sum(blk{p,2}))
            convertyes(p) = 1;
        end
    end
    if any(convertyes)
        if (par.printlevel);
            fprintf('\n sqlp: converting At into required format');
        end
        At = svec(blk,At,ones(size(blk,1),1));
    end
end
%%
%%-----------------------------------------
%% main solver
%%-----------------------------------------
%%

[obj,X,y,Z,runhist] = ...
    DSA_BD(blk,At,C,b,par,X0,y0,Z0);
end

